#ifndef MUPDF_PDF_PARSE_H
#define MUPDF_PDF_PARSE_H

/*
 * tokenizer and low-level object parser
 */

enum
{
	PDF_LEXBUF_SMALL = 256,
	PDF_LEXBUF_LARGE = 65536
};

typedef struct pdf_lexbuf_s
{
	fz_warn_context *ctx;
	unsigned int size;
	unsigned int base_size;
	unsigned int len;
	int i;
	float f;
	char *scratch;
	char buffer[PDF_LEXBUF_SMALL];
} pdf_lexbuf;

typedef struct pdf_lexbuf_large_s
{
	pdf_lexbuf base;
	char buffer[PDF_LEXBUF_LARGE - PDF_LEXBUF_SMALL];
} pdf_lexbuf_large;

typedef enum
{
	PDF_TOK_ERROR = 0, PDF_TOK_EOF = 1,
	PDF_TOK_OPEN_ARRAY, PDF_TOK_CLOSE_ARRAY,
	PDF_TOK_OPEN_DICT, PDF_TOK_CLOSE_DICT,
	PDF_TOK_OPEN_BRACE, PDF_TOK_CLOSE_BRACE,
	PDF_TOK_NAME, PDF_TOK_INT, PDF_TOK_REAL, PDF_TOK_STRING, PDF_TOK_KEYWORD,
	PDF_TOK_R, PDF_TOK_TRUE, PDF_TOK_FALSE, PDF_TOK_NULL,
	PDF_TOK_OBJ, PDF_TOK_ENDOBJ,
	PDF_TOK_STREAM, PDF_TOK_ENDSTREAM,
	PDF_TOK_XREF, PDF_TOK_TRAILER, PDF_TOK_STARTXREF,
	PDF_NUM_TOKENS
} pdf_token;

void pdf_lexbuf_init(fz_warn_context *ctx, pdf_lexbuf *lexbuf, unsigned int size);
void pdf_lexbuf_fin(pdf_lexbuf *lexbuf);
size_t pdf_lexbuf_grow(pdf_lexbuf *lexbuf);
pdf_token pdf_lex(buf_t *f, pdf_lexbuf *lexbuf);

/*
	pdf_print_token: print a lexed token to a buffer, growing if necessary
*/
void pdf_print_token(buf_t *buf, int tok, pdf_lexbuf *lex);

#endif
