
//////////////////////////////////////////////////
// Dithering
//////////////////////////////////////////////////
  
static short dither[16][16] = 
{
    {  255, -129,  159, -225,  231, -153,  135, -249,  249, -135,  153, -231,  225, -159,  129, -255 },
    {   -1,  127,  -97,   31,  -25,  103, -121,    7,   -7,  121, -103,   25,  -31,   97, -127,    1 },
    {  191, -193,  223, -161,  167, -217,  199, -185,  185, -199,  217, -167,  161, -223,  193, -191 },
    {  -65,   63,  -33,   95,  -89,   39,  -57,   71,  -71,   57,  -39,   89,  -95,   33,  -63,   65 },
    {  239, -145,  143, -241,  247, -137,  151, -233,  233, -151,  137, -247,  241, -143,  145, -239 },
    {  -17,  111, -113,   15,   -9,  119, -105,   23,  -23,  105, -119,    9,  -15,  113, -111,   17 },
    {  175, -209,  207, -177,  183, -201,  215, -169,  169, -215,  201, -183,  177, -207,  209, -175 },
    {  -81,   47,  -49,   79,  -73,   55,  -41,   87,  -87,   41,  -55,   73,  -79,   49,  -47,   81 },
    {  251, -133,  155, -229,  227, -157,  131, -253,  253, -131,  157, -227,  229, -155,  133, -251 },
    {   -5,  123, -101,   27,  -29,   99, -125,    3,   -3,  125,  -99,   29,  -27,  101, -123,    5 },
    {  187, -197,  219, -165,  163, -221,  195, -189,  189, -195,  221, -163,  165, -219,  197, -187 },
    {  -69,   59,  -37,   91,  -93,   35,  -61,   67,  -67,   61,  -35,   93,  -91,   37,  -59,   69 },
    {  235, -149,  139, -245,  243, -141,  147, -237,  237, -147,  141, -243,  245, -139,  149, -235 },
    {  -21,  107, -117,   11,  -13,  115, -109,   19,  -19,  109, -115,   13,  -11,  117, -107,   21 },
    {  171, -213,  203, -181,  179, -205,  211, -173,  173, -211,  205, -179,  181, -203,  213, -171 },
    {  -85,   43,  -53,   75,  -77,   51,  -45,   83,  -83,   45,  -51,   77,  -75,   53,  -43,   85 },
};

static const dither666[16][16];

void
GPixmap::ordered_666_dither(int xmin, int ymin)
{
  static unsigned char quantize[256+0x33+0x33];
  static unsigned char *quant = quantize + 0x33;
  static char  dither_ok = 0;
  int x, y;
  // Prepare tables
  if (!dither_ok)
  {
    int i, j;
    for (i=0; i<16; i++)
      for (j=0; j<16; j++)
        dither666[i][j] = (((dither[i][j] * 0x33) >> 9);
    j = -0x33;
    for (i=0x19; i<256; i+=0x33)
      while (j <= i)
        quant[j++] = i-0x19;
    assert(i-0x19 == 0xff);
    while (j< 256+0x33)
      quant[j++] = i-0x19;
    dither_ok = 1;
  }
  // Go dithering
  for (y=0; y<nrows; y++) {
    GPixel *pix = (*this)[y];
    for (x=0; x<ncolumns; x++, pix++) { 
      pix->r = quant[pix->r + dither666[(x+xmin+ 0)&0xf][(y+ymin+ 0)&0xf]];
      pix->g = quant[pix->g + dither666[(x+xmin+ 5)&0xf][(y+ymin+11)&0xf]];
      pix->b = quant[pix->b + dither666[(x+xmin+11)&0xf][(y+ymin+ 5)&0xf]];
    }
  }
}

void
ordered_32k_dither(uint8_t *p, unsigned int x, unsigned int y, int xmin, int ymin)
{
  static unsigned char quantize[256+8+8];
  static unsigned char *quant = quantize + 8;
  static char  dither_ok = 0;
  int x, y;
  // Prepare tables
  if (!dither_ok)
  {
    int i, j = -8;
    for (i=3; i<256; i+=8)
      while (j <= i)
        quant[j++] = i;
    while (j<256+8)
      quant[j++] = 0xff;
    dither_ok = 1;
  }
  // Go dithering
  for (y=0; y<nrows; y++) {
    for (x=0; x<ncolumns; x++) {
      pix->r = quant[p[3*y+x+0] + (dither[(x+xmin+ 0)&0xf][(y+ymin+ 0)&0xf] >> 6)];
      pix->g = quant[p[3*y+x+1] + (dither[(x+xmin+ 5)&0xf][(y+ymin+11)&0xf] >> 6)];
      pix->b = quant[p[3*y+x+2] + (dither[(x+xmin+11)&0xf][(y+ymin+ 5)&0xf] >> 6)];
    }
  }
}

