#ifndef MUPDF_FITZ_CRYPT_H
#define MUPDF_FITZ_CRYPT_H

#include "mupdf/fitz/system.h"

/*
 * Basic crypto functions.
 * Independent of the rest of fitz.
 * For further encapsulation in filters, or not.
 */

/* md5 digests */
typedef struct fz_md5_s
{
	unsigned int state[4];
	uint64_t count;
	unsigned int buffer[16];
} fz_md5;

void fz_md5_init(fz_md5 *state);
void fz_md5_update(fz_md5 *state, const unsigned char *input, unsigned inlen);
void fz_md5_final(fz_md5 *state, unsigned char digest[16]);
void fz_md5_stringify(char *buf, unsigned char *digest);

/* arc4 crypto */
typedef struct fz_arc4_s
{
	unsigned char x, y, state[256];
} fz_arc4;

static inline void fz_arc4_init(fz_arc4 *arc4, const unsigned char *key, unsigned keylen)
{
	unsigned int t, u, keyindex = 0, stateindex = 0;
	unsigned char *state = arc4->state;
	unsigned int counter;
	arc4->x = 0;
	arc4->y = 0;

	for (counter = 0; counter < 256; counter++) {
		state[counter] = counter;
	}

	for (counter = 0; counter < 256; counter++) {
		t = state[counter];
		stateindex = (stateindex + key[keyindex] + t) & 0xff;
		u = state[stateindex];
		state[stateindex] = t;
		state[counter] = u;
		if (++keyindex >= keylen) {
			keyindex = 0;
		}
	}
}

static inline unsigned char
fz_arc4_next(fz_arc4 *arc4)
{
	unsigned int y = arc4->y, sx, sy;
	unsigned char *state = arc4->state;
	sx = state[++arc4->x];
	y = (sx + y) & 0xff;
	sy = state[y];
	arc4->y = y;
	state[arc4->y] = sx;
	state[arc4->x] = sy;
	return state[(sx + sy) & 0xff];
}

/* AES block cipher implementation from XYSSL */
typedef struct fz_aes_s {
    unsigned int  eK[68], dK[68]; /*!<  AES round keys    */
    unsigned int  Nr;             /*!<  number of rounds  */
    unsigned char aesni;          /*!<  AES-NI supported? */
    unsigned char iv_nonce[16];
    unsigned int  iv_offset;
} fz_aes;

int aes_setup(fz_aes *ctx, const unsigned char *key, unsigned int keysize);
void aes_ecb_encrypt(fz_aes *ctx, const unsigned char *input, unsigned char *output);
void aes_ecb_decrypt(fz_aes *ctx, const unsigned char *input, unsigned char *output);

#endif
