#ifndef MUPDF_PDF_RESOURCE_H
#define MUPDF_PDF_RESOURCE_H

/*
 * PDF interface to store
 */
void pdf_store_item(fz_context *ctx, pdf_obj *key, void *val, unsigned int itemsize);
void *pdf_find_item(fz_context *ctx, fz_store_free_fn *free, pdf_obj *key);
void pdf_remove_item(fz_context *ctx, fz_store_free_fn *free, pdf_obj *key);

/*
 * PostScript calculator
 */
typedef struct psobj_s psobj;
typedef struct pdf_function_s pdf_function;
void pdf_eval_postscript_func(fz_context *ctx, pdf_function *func, float *in, float *out);
int pdf_load_postscript_func(pdf_function *func, pdf_document *doc, pdf_obj *dict, int num, int gen);
void pdf_debug_postscript_func(psobj *funccode, psobj *code, int level);

/*
 * Functions, Colorspaces, Shadings and Images
 */

enum
{
	SAMPLE = 0,
	EXPONENTIAL = 2,
	STITCHING = 3,
	POSTSCRIPT = 4
};

//fz_colorspace *pdf_load_colorspace(pdf_document *doc, pdf_obj *obj);
//fz_shade *pdf_load_shading(pdf_document *doc, pdf_obj *obj);

/*
 * Pattern
 */

typedef struct pdf_pattern_s pdf_pattern;

struct pdf_pattern_s
{
	fz_storable storable;
	int ismask;
	float xstep;
	float ystep;
	fz_matrix matrix;
	fz_rect bbox;
	pdf_obj *resources;
	pdf_obj *contents;
};

//pdf_pattern *pdf_load_pattern(pdf_document *doc, pdf_obj *obj);
//pdf_pattern *pdf_keep_pattern(fz_context *ctx, pdf_pattern *pat);
//void pdf_drop_pattern(fz_context *ctx, pdf_pattern *pat);

/*
 * XObject
 */

typedef struct pdf_xobject_s pdf_xobject;

struct pdf_xobject_s
{
	fz_storable storable;
	fz_matrix matrix;
	fz_rect bbox;
	int isolated;
	int knockout;
	int transparency;
	fz_colorspace *colorspace;
	pdf_obj *resources;
	pdf_obj *contents;
	pdf_obj *me;
	int iteration;
};

void pdf_update_appearance(fz_context *ctx, pdf_document *doc, pdf_annot *annot);

#endif
