#ifndef MUPDF_PDF_XREF_H
#define MUPDF_PDF_XREF_H

/*
	pdf_create_object: Allocate a slot in the xref table and return a fresh unused object number.
*/
int pdf_create_object(pdf_document *doc);

/*
	pdf_delete_object: Remove object from xref table, marking the slot as free.
*/
void pdf_delete_object(pdf_document *doc, int num);

/*
	pdf_update_object: Replace object in xref table with the passed in object.
*/
void pdf_update_object(pdf_document *doc, int num, pdf_obj *obj);

/*
	pdf_update_stream: Replace stream contents for object in xref table with the passed in buffer.

    The buffer contents must match the /Filter setting if 'compressed' is true.
    If 'compressed' is false, the /Filter and /DecodeParms entries are deleted.
    The /Length entry is updated.
 */
void pdf_update_stream(pdf_document *doc, pdf_obj *ref, fz_buffer *newbuf, int compressed);

/*
 * xref and object / stream api
 */

typedef struct pdf_xref_entry_s pdf_xref_entry;

struct pdf_xref_entry_s
{
	char type;	/* 0=unset (f)ree i(n)use (o)bjstm */
	unsigned char flags; /* bit 0 = marked */
	unsigned short gen;	/* generation / objstm index */
	int ofs;	/* file offset / objstm object number */
	int stm_ofs;	/* on-disk stream */
	fz_buffer *stm_buf; /* in-memory stream (for updated objects) */
	pdf_obj *obj;	/* stored/cached object */
};

struct pdf_xref_s
{
	int len;
	pdf_xref_entry *table;
	pdf_obj *trailer;
};

#define PDF_OBJ_FLAG_MARK 1

void pdf_cache_object(pdf_document *doc, unsigned int num, int gen);

int pdf_count_objects(pdf_document *doc);
pdf_obj *pdf_resolve_indirect(pdf_obj *ref);
pdf_obj *pdf_load_object(pdf_document *doc, int num, int gen);

//fz_buffer *pdf_load_raw_stream(pdf_document *doc, int num, int gen);
//fz_stream *pdf_open_raw_stream(pdf_document *doc, int num, int gen);
fz_buffer *pdf_load_stream(pdf_document *doc, int num, int gen);
fz_stream *pdf_open_stream(pdf_document *doc, int num, int gen);

fz_stream *pdf_open_inline_stream(pdf_document *doc, pdf_obj *stmobj, int length, fz_stream *chain, fz_compression_params *params);
fz_compressed_buffer *pdf_load_compressed_stream(pdf_document *doc, int num, int gen);
fz_stream *pdf_open_stream_with_offset(pdf_document *doc, int num, int gen, pdf_obj *dict, int stm_ofs);
fz_stream *pdf_open_compressed_stream(fz_context *ctx, fz_compressed_buffer *);
fz_stream *pdf_open_contents_stream(pdf_document *doc, pdf_obj *obj);
fz_buffer *pdf_load_raw_renumbered_stream(pdf_document *doc, int num, int gen, int orig_num, int orig_gen);
fz_buffer *pdf_load_renumbered_stream(pdf_document *doc, int num, int gen, int orig_num, int orig_gen, int *truncated);
fz_stream *pdf_open_raw_renumbered_stream(pdf_document *doc, int num, int gen, int orig_num, int orig_gen);

static inline fz_stream *pdf_open_raw_stream(pdf_document *doc, int num, int gen)
{
	return pdf_open_raw_renumbered_stream(doc, num, gen, num, gen);
}

static inline fz_buffer *pdf_load_raw_stream(pdf_document *doc, int num, int gen)
{
	return pdf_load_raw_renumbered_stream(doc, num, gen, num, gen);
}

pdf_obj *pdf_trailer(pdf_document *doc);
void pdf_set_populating_xref_trailer(pdf_document *doc, pdf_obj *trailer);
int pdf_xref_len(pdf_document *doc);
pdf_xref_entry *pdf_get_populating_xref_entry(pdf_document *doc, int i);
pdf_xref_entry *pdf_get_xref_entry(pdf_document *doc, int i);
void pdf_replace_xref(pdf_document *doc, pdf_xref_entry *entries, int n);
void pdf_xref_ensure_incremental_object(pdf_document *doc, int num);
int pdf_xref_is_incremental(pdf_document *doc, int num);

void pdf_mark_xref(pdf_document *doc);
void pdf_clear_xref(pdf_document *doc);
void pdf_clear_xref_to_mark(pdf_document *doc);

void pdf_repair_xref(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf);
void pdf_repair_obj_stms(fz_context *ctx, pdf_document *doc);
pdf_obj *pdf_new_ref(pdf_document *doc, pdf_obj *obj);

int pdf_repair_obj(fz_context *ctx, pdf_document *doc, pdf_lexbuf *buf, int *stmofsp, int *stmlenp, pdf_obj **encrypt, pdf_obj **id, pdf_obj **page, int *tmpofs);

pdf_obj *pdf_progressive_advance(pdf_document *doc, int pagenum);

void pdf_bufprint_xref(fz_buffer *outbuf, pdf_document *doc);

#endif
