#ifndef MUPDF_PDF_CMAP_H
#define MUPDF_PDF_CMAP_H

/*
 * CMap
 */

typedef struct pdf_cmap_s pdf_cmap;

enum { PDF_CMAP_SINGLE, PDF_CMAP_RANGE, PDF_CMAP_TABLE };
#define PDF_MRANGE_CAP 8
#define PDF_NUM_CODESPACES 40

typedef struct pdf_range_s
{
	unsigned short low;
	/* Next, we pack 2 fields into the same unsigned short. Top 14 bits
	 * are the extent, bottom 2 bits are flags: single, range, table */
	unsigned short extent_flags;
	unsigned short out;	/* range-delta or table-index */
} pdf_range;

typedef struct pdf_xrange_s
{
	unsigned int low, high, out;
} pdf_xrange;

struct pdf_cmap_s
{
	char cmap_name[32];
	char usecmap_name[32];
	pdf_cmap *usecmap;

	int wmode;

	int codespace_len;
	struct
	{
		unsigned short n;
		unsigned short low;
		unsigned short high;
	} codespace[40];

	unsigned int rlen, rcap;
	pdf_range *ranges;

	unsigned int tlen, tcap;
	unsigned short *table;

	unsigned int xlen, xcap;
	pdf_xrange *xranges;
};

typedef struct fz_warn_context_s
{
	char message[1024];
	unsigned int msglen, count;
} fz_warn_context;

void fz_warn(fz_warn_context *ctx, const char *fmt, ...);
void fz_flush_warnings(fz_warn_context *ctx);
size_t pdf_strlcpy(char *dst, const char *src, size_t siz);

pdf_cmap *pdf_new_cmap(void);
void pdf_free_cmap(pdf_cmap *cmap);
unsigned int pdf_cmap_size(pdf_cmap *cmap);

int pdf_cmap_wmode(pdf_cmap *cmap);
void pdf_set_cmap_wmode(pdf_cmap *cmap, int wmode);

void pdf_add_codespace(fz_warn_context *warn, pdf_cmap *cmap, int low, int high, int n);
void pdf_map_range_to_table(pdf_cmap *cmap, int low, int *map, int len);
void pdf_map_range_to_range(pdf_cmap *cmap, int srclo, int srchi, int dstlo);
void pdf_map_one_to_many(pdf_cmap *cmap, int one, int *many, int len);
void pdf_sort_cmap(fz_warn_context *ctx, pdf_cmap *cmap);

int pdf_lookup_cmap(pdf_cmap *cmap, int cpt);
unsigned int pdf_decode_cmap(pdf_cmap *cmap, unsigned char *buf, unsigned int len, unsigned int *cpt);

pdf_cmap *pdf_new_identity_cmap(fz_warn_context *warn, int wmode, int bytes);

#ifndef NDEBUG
void pdf_print_cmap(pdf_cmap *cmap);
#endif

#endif
