/* fontdump.c -- an "xxd -i" workalike for dumping binary fonts as source code */
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

static int
hexdump(FILE *fo, unsigned char *inbuf, size_t insize)
{
    int c;
    size_t k = 0;
    while (k < insize) {
        c = inbuf[k];
        fprintf(fo, "0x%02x,", c);
        if ((k&15) == 15) {
            fprintf(fo, "\n  ");
        }
        k++;
    }
    fprintf(fo, "\n");
    return k;
}

int
main(int argc, char **argv)
{
    struct stat st;
    FILE *fo;
    int fi = -1;
    unsigned char *inbuf = NULL;
    char fontname[256];
    char origname[256];
    char *basename;
    char *p;
    unsigned int i, len = 0;

    if (argc < 3) {
        fprintf(stderr, "usage: fontdump output.c input.dat\n");
        return 1;
    }

    fo = fopen(argv[1], "wb");
    if (!fo) {
        fprintf(stderr, "fontdump: could not open output file '%s'\n", argv[1]);
        return 1;
    }

    for (i = 2; i < argc; i++) {
        if ((fi = open(argv[i], O_RDONLY)) < 0) {
            fclose(fo);
            fprintf(stderr, "fontdump: could not open input file '%s'\n", argv[i]);
            return 1;
        }

        basename = strrchr(argv[i], '/');
        if (!basename)
            basename = strrchr(argv[i], '\\');
        if (basename)
            basename++;
        else
            basename = argv[i];

        strcpy(origname, basename);
        p = strrchr(origname, '.');
        if (p) *p = 0;
        strcpy(fontname, origname);

        p = fontname;
        while (*p) {
            if (*p == '/' || *p == '.' || *p == '\\' || *p == '-')
                *p = '_';
            p ++;
        }

        fstat(fi, &st);
        len = st.st_size;
        inbuf = malloc(len+1);
        read(fi, inbuf, len);
        close(fi);

        fprintf(stdout, "\t{\"%s\",pdf_font_%s,%u},\n", origname, fontname, len);
        fprintf(fo, "static const unsigned char pdf_font_%s[%u] = {\n  ", fontname, len);
        hexdump(fo, inbuf, len);
        fprintf(fo, "};\n");
    }

    if (fclose(fo)) {
        fprintf(stderr, "fontdump: could not close output file '%s'\n", argv[1]);
        return 1;
    }

    return 0;
}

