#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <float.h>
#include <sys/time.h>
#include <xmmintrin.h>
uint32_t fz_ftoa(char *buf, float f, int *expt);

typedef struct {
    unsigned int state[64];
    unsigned int index;                                                                                                                                 
} AVLFG;

static void lfg_srand(AVLFG *c, unsigned int seed){
    uint32_t i, tmp = seed;

    for(i=0; i<64; i+=1){
        tmp = ((tmp * 1664525) + 1013904223);
        c->state[i]= tmp;
    }
    c->index=0;
}

static unsigned int lfg_rand(AVLFG *c){
    c->state[c->index & 63] = c->state[(c->index-24) & 63] + c->state[(c->index-55) & 63];
    return c->state[c->index++ & 63];
}

int main(void) {
    char buf1[1024], buf2[1024];
    AVLFG c;
    unsigned int i, j;
    float f, g;
    struct timeval tv;
    gettimeofday(&tv, NULL);
    lfg_srand(&c, 0xDEADBEEF);
    setvbuf(stdout, NULL, _IONBF, 0);
    _mm_setcsr(_mm_getcsr() | (1<<15) | (1<<6));
    for (i = 0; i < 65536; i++) {
        int expt = 0;
        g = lfg_rand(&c);
        f = lfg_rand(&c);
        f /= g;
        f *= (float)(lfg_rand(&c) & 0xffff);
        f *= 3.1415926535897f;
        j = fz_ftoa(buf1, f, NULL);
        snprintf(buf2, 1023, "%.6f", f);
        if (strcmp(buf1, buf2))
            printf("%u: mismatch: %s %s\n", i, buf1, buf2);
        f  = lfg_rand(&c);
        f /= (float)lfg_rand(&c);
        f /= g;
        f *= 3.1415926535897f;
        j = fz_ftoa(buf1, f, &expt);
        buf1[j] = 0;
        snprintf(buf2, 1023, "%.16f", f);
        if (1||memcmp(buf1, buf2, 7))
            printf("%u: mismatch: %s %s\n", i, buf1, buf2);
        f  = lfg_rand(&c);
        f /= (float)(lfg_rand(&c) & 0xFF);
        f /= g;
        f *= 3.1415926535897f;
        j = fz_ftoa(buf1, f, &expt);
        snprintf(buf2, 1023, "%.6f", f);
        if (1||memcmp(buf1, buf2, 7))
            printf("%u: mismatch: %s %s\n", i, buf1, buf2);
    }
#if 0
    for (i = 0; i < 256; i++) {
        f = ((float)i / 65536.0f);
        f *= 3.1415926535897f;
        f *= 3.1415926535897f;
        j = fz_ftoa(buf1, f, 7, 0, NULL);
        printf("%.7f %s\n", f, buf1);
    }
    for (i = 0; i < 32; i++) {
        f = ((float)i / 32.0f);
        f *= 3.1415926535897f;
        f += 9.0f;
        j = fz_ftoa(buf1, f, 7, 0, NULL);
        printf("%.7f %s\n", f, buf1);
    }
    _f32 f;
    unsigned int i, j, len1, len2, mismatches = 0;
    for (j = 0; j < 128; j++) {
        for (i = 0; i < (1 << 23U); i++) {
            f.i = ((0x3f8 + j) << 20)|i;
            len1 = fz_ftoa(buf1, f.f, 6, 'f');
            len2 = snprintf(buf2, 1023, "%.6f", f.f);
            len1--; len2--;
            while(buf2[len2] == '0') buf2[len2--] = '\0';
            len2++;
            if (strcmp(buf1, buf2)) {
                mismatches++;
#if 1
                printf("mismatch: buf1: %s != buf2: %s (i: 0x%06x/0x%02x/0x%02x, j: %u, len1: %u, len2: %u)\n",
                       buf1, buf2, i, (i>>13)&0x0c, (i>>13)&0x03, j, len1, len2);
#endif      
            }
        }
        printf("total mismatched roundings: %u (probability: %.9f)\n", mismatches, (double)mismatches / (double)(1 << 23U));
        mismatches = 0;
    }
#endif
    return 0;
}

