/*
 * This is a version of the public domain getopt implementation by
 * Henry Spencer originally posted to net.sources.
 *
 * This file is in the public domain.
 */

#if defined(_WIN32) || defined(_WIN64)

#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include "mupdf/fitz/getopt.h"

#define wgetopt fz_wgetopt
#define woptarg fz_woptarg
#define woptind fz_woptind

uint16_t *woptarg; /* Global argument pointer. */
int woptind = 0; /* Global argv index. */

static uint16_t *scan = NULL; /* Private scan pointer. */

int wgetopt(int argc, uint16_t *argv[], const uint16_t *optstring)
{
	uint16_t c, *place;

	woptarg = NULL;

	if (!scan || *scan == L'\0') {
		if (woptind == 0)
			woptind++;
		if (woptind >= argc || argv[woptind][0] != L'-' || argv[woptind][1] == L'\0')
			return EOF;
		if (argv[woptind][1] == L'-' && argv[woptind][2] == L'\0') {
			woptind++;
			return EOF;
		}
		scan = argv[woptind]+1;
		woptind++;
	}

	c = *scan++;
	place = wcschr(optstring, c);
	if (!place || c == L':') {
		return L'?';
	}

	place++;
	if (*place == L':') {
		if (*scan != L'\0') {
			woptarg = scan;
			scan = NULL;
		} else if(woptind < argc) {
			woptarg = argv[woptind++];
		} else {
			return L':';
		}
	}

	return c;
}

#endif

