#include "mupdf/fitz.h"

void
fz_free_link_dest(fz_context *ctx, fz_link_dest *dest)
{
	switch (dest->kind)
	{
	case FZ_LINK_NONE:
	case FZ_LINK_GOTO:
		break;
	case FZ_LINK_URI:
		fz_free(ctx, dest->ld.uri.uri);
		break;
	case FZ_LINK_LAUNCH:
		fz_free(ctx, dest->ld.launch.file_spec);
		break;
	case FZ_LINK_NAMED:
		fz_free(ctx, dest->ld.named.named);
		break;
	case FZ_LINK_GOTOR:
		fz_free(ctx, dest->ld.gotor.file_spec);
		break;
	}
}

void
fz_free_outline(fz_context *ctx, fz_outline *outline)
{
	while (outline)
	{
		fz_outline *next = outline->next;
		fz_free_outline(ctx, outline->down);
		fz_free(ctx, outline->title);
		fz_free_link_dest(ctx, &outline->dest);
		fz_free(ctx, outline);
		outline = next;
	}
}

static void
do_debug_outline_xml(fz_buffer *buf, fz_outline *outline, int level)
{
	while (outline)
	{
		fz_buffer_printf(buf, "<outline title=\"%s\" page=\"%d\"", outline->title, outline->dest.kind == FZ_LINK_GOTO ? outline->dest.ld.gotor.page + 1 : 0);
		if (outline->down) {
			fz_write_buffer(buf, ">\n", 2);
			do_debug_outline_xml(buf, outline->down, level + 1);
			fz_write_buffer(buf, "</outline>\n", 11);
		} else {
			fz_write_buffer(buf, " />\n", 4);
		}
		outline = outline->next;
	}
}

void
fz_print_outline_xml(fz_buffer *buf, fz_outline *outline)
{
	do_debug_outline_xml(buf, outline, 0);
}

static void
do_debug_outline(fz_buffer *buf, fz_outline *outline, int level)
{
	int i;
	while (outline)
	{
		for (i = 0; i < level; i++)
			fz_write_buffer_byte(buf, '\t');
		fz_buffer_printf(buf, "%s\t%d\n", outline->title, outline->dest.kind == FZ_LINK_GOTO ? outline->dest.ld.gotor.page + 1 : 0);
		if (outline->down)
			do_debug_outline(buf, outline->down, level + 1);
		outline = outline->next;
	}
}

void
fz_print_outline(fz_buffer *buf, fz_outline *outline)
{
	do_debug_outline(buf, outline, 0);
}

