#include <stdint.h>
#include <string.h>
#include <math.h>
#include "mupdf/fitz.h"
unsigned int fz_ftoa(char *s, float v, int* K);

typedef union _f32 {
    float f;
    uint32_t i;
} _f32;

static __attribute__((always_inline)) inline int fft_ilogbf(float x)
{
    _f32 hx;
    hx.f = fabsf(x);
    return ((hx.i >> 23) - 127);
}

unsigned int fft_itoa10(char *__restrict _buf, uint32_t uval)
{
    unsigned int i = 0, j = 0;
    if (!_buf) return 25;
    do {
        _buf[i++] = ((uint8_t)(uval % 10) + 0x30);
        uval = (uval / 10);
    } while (uval);
    while(j < (i>>1)) {
        char ch = _buf[i-1-j];
        _buf[i-1-j] = _buf[j];
        _buf[j++] = ch;
    }
    _buf[i] = '\0';
    return i;
}

unsigned int fft_itoa16(char *__restrict _buf, uint32_t uval)
{
    static const char hex[16] = "0123456789abcdef";
    unsigned int i = 0, j = 0;
    if (!_buf) return 25;
    do {
        _buf[i++] = hex[(uval & 15)];
        uval = (uval >> 4);
    } while (uval);
    while(j < (i>>1)) {
        char ch = _buf[i-1-j];
        _buf[i-1-j] = _buf[j];
        _buf[j++] = ch;
    }
    _buf[i] = '\0';
    return i;
}

/**********************************************************************/
#define PA_FLAG_MASK        (0xff00)
#define PA_FLAG_LONG        (0x0400)
#define PA_FLAG_LONG_LONG   (0x0800)
#define __PA_INTMASK        (0x0f00) /* non-gnu -- all int flags */
/* We assume int may be short or long, but short and long are different. */
static inline uintmax_t _load_inttype(unsigned int desttype, register const void *src)
{
    if (desttype & (PA_FLAG_LONG|PA_FLAG_LONG_LONG)) {
        unsigned long int x = *((unsigned long int *) src);
        return x;
    } else {
        unsigned int x = *((unsigned int *) src);
        return x;
    }
}

/**********************************************************************/
/* layout                     012345678901234567   */
static const char u_spec[] = "%npxXudicCsfgMR";

#define PUTC(C) do {        \
        if (buf && (idx < lim))      \
            buf[idx] = (C); \
        idx++;              \
    } while(0)

static unsigned int fmtfloat(char *buf, float f, unsigned int idx, unsigned int len)
{
    int k = fft_ilogbf(f);
    unsigned int l = 891; 
    if (buf)
	    l = fz_ftoa(buf+idx, f, &k);
    return l;
}

int fz_vsnprintf(char *buf, size_t lim, register const char * __restrict fmt, va_list ap)
{
	int i, j, cnt = 0, dataargtype = 0, z;
	register char *p;
	const char *fmt0;
    size_t idx = 0;
	int preci, dpoint, flag = 0;

	while (*fmt) {
		if (*fmt == '%') {
			fmt0 = fmt;			/* save our position in case of bad format */
			++fmt;
			preci = 5;			/* max string width or mininum digits */
			z = dpoint = 0;		/* found decimal point */

			/* process optional flags */
			if (*fmt == ' ' || *fmt == '+') flag = *fmt++;
			if (*fmt == '0' || *fmt == '-') { fmt++; z = 1; }
            if (*fmt == '#') fmt++;

			/* process optional width and precision */
			do {
				if (*fmt == '.') {
					++fmt;
					dpoint = 1;
				}
				if (*fmt == '*') {	/* parameter width value */
					++fmt;
					i = va_arg(ap, int);
				} else {
					for ( i = 0 ; (*fmt >= '0') && (*fmt <= '9') ; ++fmt ) {
						i = (i * 10) + (*fmt - '0');
					}
				}

				if (dpoint) {
					preci = i;
					if (i<0) { preci = 5; }
				} else if (z) {
				    z = i;
                }
			} while ((*fmt == '.') && !dpoint);

			/* process optional qualifier */
            dataargtype = 0;
            if (*fmt == 'z' || *fmt == 'j' || *fmt == 't' || *fmt == 'l') {
                fmt++;
                dataargtype = PA_FLAG_LONG;
            }
            if (*fmt == 'Z' || *fmt == 'q' || *fmt == 'L' || *fmt == 'l') {
                fmt++;
                dataargtype = PA_FLAG_LONG_LONG;
            }

			/* process format specifier */
			for (p = (char *) u_spec ; *p ; p++) {
				if (*fmt != *p) continue;
				if (p-u_spec < 1) {	/* print a % */
		            PUTC('%'); /* normal char out */
					goto nextfmt;
				}
				if (p-u_spec < 2) {	/* store output count in int ptr */
					goto nextfmt;
				}

				if (*p == 'p') {
                    dataargtype = PA_FLAG_LONG_LONG;
				}

				if (p-u_spec < 8) { /* (un)signed conversion */
					unsigned int radix = (((*p|0x20) == 'x') ? 16 : 10);
                    unsigned long k = ((dataargtype & (PA_FLAG_LONG|PA_FLAG_LONG_LONG)) ? va_arg(ap, long) : va_arg(ap, int));
					if (*p == 'p') {
                        if (!k) { /* null pointer */
                            if (buf && (idx+5 < lim))
						        memcpy(buf+idx, "(nil)", 5);
                            idx += 5;
					        goto nextfmt;
                        }
					}
                    if(!buf || idx+25 >= lim) {
                        idx += 25;
                    } else {
                        if(radix == 10) {
                            if (p-u_spec > 5) {
                                if(k > INTMAX_MAX) {
                                    k = -k;
                                    PUTC('-');
                                } else if((!(dataargtype&0x0C00)) && (k > INT32_MAX)) {
                                    k = -(uint32_t)k;
                                    PUTC('-');
                                } else if (flag) {
                                    PUTC(flag);
                                }
                            }
                            idx += fft_itoa10(buf+idx, k);
                        } else {
                            PUTC('0');
                            PUTC(*fmt);
                            idx += fft_itoa16(buf+idx, k);
                        }
                    }
					goto nextfmt;
				} else if (p-u_spec < 11) {	/* character or string */
					if (*p == 'c') {	/* character */
						char ch = va_arg(ap, int);
						/* This takes care of the "%c",0 case */
                        PUTC(ch);
					} else if (*p == 'C') {	/* character */
						unsigned int ch = va_arg(ap, int);
				        if (ch < 128) {
                            PUTC(ch);
				        } else {
					        char utf8buf[7];
					        unsigned int j = fz_runetochar(utf8buf, ch);
                            if (buf && (idx+j < lim))
                                memcpy(buf+idx, utf8buf, j);
                            idx += j;
				        }
					} else {	/* string */
						p = va_arg(ap, char *);
						if (!p) {
                            if (buf && (idx+6 < lim))
						        memcpy(buf+idx, "(null)", 6);
                            idx += 6;
						}
                        j = strlen(p);
                        if (buf && (idx+j < lim))
                            memcpy(buf+idx, p, j);
                        idx += j;
                        while (j++ < z) {
                            PUTC(' ');
                        }
					}
					goto nextfmt;
				} else if (p-u_spec < 13) { /* floating point */
                    float f = (float)va_arg(ap, double);
                    if (f < 0.0f) { flag = '-'; f = -f; }
                    if (flag) PUTC(flag);
                    idx += fmtfloat(buf, f, idx, lim);
					goto nextfmt;
				} else if (p-u_spec < 14) { /* fz_matrix */
				    fz_matrix *m = va_arg(ap, fz_matrix*);
                    unsigned int k;
                    for (k = 0; k < 4; k++) {
                        if (m->x[0] < 0.0f) { flag = '-'; m->x[0] = -m->x[0]; } if (flag) PUTC(flag);
                        idx += fmtfloat(buf, m->x[0], idx, lim); PUTC(' ');
                    }
                    if (m->e < 0.0f) { flag = '-'; m->e = -m->e; } if (flag) PUTC(flag);
                    idx += fmtfloat(buf, m->e, idx, lim); PUTC(' ');
                    if (m->f < 0.0f) { flag = '-'; m->f = -m->f; } if (flag) PUTC(flag);
                    idx += fmtfloat(buf, m->f, idx, lim);
                }
			}

			fmt = fmt0;	/* this was an illegal format */
		}
		++cnt;
		PUTC(*fmt); /* normal char out */

nextfmt:
		++fmt;
	}
	return idx;
}

int fz_snprintf(char *str, size_t size, const char * __restrict format, ...)
{
    va_list arg;
    int rv;
    va_start(arg, format);
    rv = fz_vsnprintf(str, size, format, arg);
    va_end(arg);
    return rv;
}

