#include "mupdf/fitz.h"

#if defined(_WIN32) && !defined(NDEBUG)
#define USE_OUTPUT_DEBUG_STRING
#endif

#ifdef USE_OUTPUT_DEBUG_STRING
#include <windows.h>
#endif

/* Warning context */

void fz_flush_warnings(fz_warn_context *ctx)
{
	if (ctx->count > 1) {
        char tmpbuf[1024];
		int len = snprintf(tmpbuf, 1023, "warning: ... repeated %u times ...\n", ctx->count);
        write(2, tmpbuf, len);
		LOGE("warning: ... repeated %d times ...\n", ctx->count);
	}
	ctx->message[0] = 0;
    ctx->msglen = 0;
	ctx->count = 0;
}

void fz_warn(fz_warn_context *ctx, const char *fmt, ...)
{
	va_list ap;
	char buf[1024];
    size_t buflen;

	va_start(ap, fmt);
	buflen = vsnprintf(buf, sizeof(buf), fmt, ap);
	va_end(ap);
#ifdef USE_OUTPUT_DEBUG_STRING
	OutputDebugStringA(buf);
	OutputDebugStringA("\n");
#endif

    if (!ctx) {
        write(2, "warning: ", 9);
        write(2, buf, buflen);
        write(2, "\n", 1);
		LOGE("warning: %s\n", buf);
    } else {
	    if ((buflen == ctx->msglen) && !memcmp(buf, ctx->message, ctx->msglen)) {
		    ctx->count++;
	    } else {
		    fz_flush_warnings(ctx);
            write(2, "warning: ", 9);
            write(2, buf, buflen);
            write(2, "\n", 1);
		    LOGE("warning: %s\n", buf);
		    memcpy(ctx->message, buf, buflen+1);
            ctx->msglen = buflen;
		    ctx->count = 1;
	    }
    }
}

#include <ft2build.h>
/* concatenate C tokens */
#define FT_ERR_XCAT( x, y )  x ## y
#define FT_ERR_CAT( x, y )   FT_ERR_XCAT( x, y )
#define FT_ERR( e )  FT_ERR_CAT( FT_ERR_PREFIX, e )
#include FT_ERRORS_H

#undef __FTERRORS_H__
#define FT_ERRORDEF(e, v, s) { (e), (s) },
#define FT_ERROR_START_LIST
#define FT_ERROR_END_LIST { 0, NULL }

struct ft_error
{
	int err;
	char *str;
};

static const struct ft_error ft_errors[] =
{
#include FT_ERRORS_H
};

char *ft_error_string(int err)
{
	const struct ft_error *e;

	for (e = ft_errors; e->str; e++)
		if (e->err == err)
			return e->str;

	return "Unknown error";
}

