#include "mupdf/fitz.h"
#include <zlib.h>
#define IHDR_TAG 0x52444849
#define IDAT_TAG 0x54414449
#define IEND_TAG 0x444E4549

static __attribute__((always_inline)) inline uint32_t av_bswap32(uint32_t x){
    __asm__("bswap   %0" : "+r" (x));
    return x;
}

/* NB: width, height and crc32 are all *big*-endian!!! */
typedef struct PNGHeader {
    uint32_t width, height;
    uint8_t bit_depth, color_type, compression_type,
            filter_type, interlace_type;
} __attribute__((packed)) PNGHeader;

struct info
{
	unsigned int width, height, depth, n;
	int interlace, indexed;
	unsigned int size;
	unsigned char *samples;
	unsigned char palette[256*4];
	int transparency;
	int trns[3];
	int xres, yres;
};

static inline unsigned int getuint(unsigned char *p)
{
    unsigned int *p32 = (unsigned int*)p;
	return av_bswap32(p32[0]);
}

static inline int getcomp(unsigned char *line, int x, int bpc)
{
	switch (bpc)
	{
	case 1: return (line[x >> 3] >> ( 7 - (x & 7) ) ) & 1;
	case 2: return (line[x >> 2] >> ( ( 3 - (x & 3) ) << 1 ) ) & 3;
	case 4: return (line[x >> 1] >> ( ( 1 - (x & 1) ) << 2 ) ) & 15;
	case 8: return line[x];
	case 16: return line[x << 1] << 8 | line[(x << 1) + 1];
	}
	return 0;
}

static inline void putcomp(unsigned char *line, int x, int bpc, int value)
{
	int maxval = (1 << bpc) - 1;

	switch (bpc)
	{
	case 1: line[x >> 3] &= ~(maxval << (7 - (x & 7))); break;
	case 2: line[x >> 2] &= ~(maxval << ((3 - (x & 3)) << 1)); break;
	case 4: line[x >> 1] &= ~(maxval << ((1 - (x & 1)) << 2)); break;
	}

	switch (bpc)
	{
	case 1: line[x >> 3] |= value << (7 - (x & 7)); break;
	case 2: line[x >> 2] |= value << ((3 - (x & 3)) << 1); break;
	case 4: line[x >> 1] |= value << ((1 - (x & 1)) << 2); break;
	case 8: line[x] = value; break;
	case 16: line[x << 1] = value >> 8; line[(x << 1) + 1] = value & 0xFF; break;
	}
}

static inline int paeth(int a, int b, int c)
{
	/* The definitions of ac and bc are correct, not a typo. */
	int ac = b - c, bc = a - c, abcc = ac + bc;
	int pa = (ac < 0 ? -ac : ac);
	int pb = (bc < 0 ? -bc : bc);
	int pc = (abcc < 0 ? -abcc : abcc);
	return pa <= pb && pa <= pc ? a : pb <= pc ? b : c;
}

static void add_bytes_l2_mmx(uint8_t *dst, uint8_t *src1, uint8_t *src2, int w)
{
    uint64_t i=0;
    __asm__ volatile(
        "jmp 2f                         \n\t"
        "1:                             \n\t"
        "movq   (%2, %0), %%mm0         \n\t"
        "movq  8(%2, %0), %%mm1         \n\t"
        "paddb  (%3, %0), %%mm0         \n\t"
        "paddb 8(%3, %0), %%mm1         \n\t"
        "movq %%mm0,  (%1, %0)          \n\t"
        "movq %%mm1, 8(%1, %0)          \n\t"
        "add $16, %0                    \n\t"
        "2:                             \n\t"
        "cmp %4, %0                     \n\t"
        "js 1b                          \n\t"
        : "+r" (i)
        : "r"(dst), "r"(src1), "r"(src2), "r"((uint64_t)w-15)
    );
    for(; i<w; i++)
        dst[i] = src1[i] + src2[i];
}

static void
png_predict(unsigned char *samples, unsigned int width, unsigned int height, unsigned int n, unsigned int depth)
{
	unsigned int stride = (width * n * depth + 7) / 8;
	unsigned int bpp = (n * depth + 7) / 8;
	unsigned int i, row;

	for (row = 0; row < height; row ++) {
		unsigned char *src = samples + (unsigned int)((stride + 1) * row);
		unsigned char *prev = samples + (unsigned int)((stride - 1) * row);
		unsigned char *dst = samples + (unsigned int)(stride * row);
        unsigned char filter_type = *src++;

        if (row == 0) {
		    switch (filter_type)
	    	{
	    	    default:
		        case 0: /* None */
		        case 2: /* Up */
                    memcpy(dst, src, stride);
			        break;
		        case 1: /* Sub */
		        case 4: /* Paeth */
			        for (i = 0; i < bpp; i++)
				        dst[i] = src[i];
			        for (i = bpp; i < stride; i++)
				        dst[i] += src[i];
			        break;
		        case 3: /* Average */
				    for (i = 0; i < bpp; i++)
					    dst[i] = src[i];
				    for (i = bpp; i < stride; i++)
					    dst[i] = src[i] + (dst[i] >> 1);
			        break;
            }
        } else {
    		switch (filter_type)
    		{
    		    default:
    		    case 0: /* None */
                    memcpy(dst, src, stride);
    			    break;
    
    		    case 1: /* Sub */
    			    for (i = 0; i < bpp; i++)
    				    dst[i] = src[i];
    			    for (i = bpp; i < stride; i++)
    				    dst[i] += src[i];
    			    break;
    
    		    case 2: /* Up */
    				add_bytes_l2_mmx(dst, src, prev, stride);
    			    break;
    
    		    case 3: /* Average */
    				for (i = 0; i < bpp; i++)
    					dst[i] = src[i] + (prev[i] >> 1);
    				for (i = bpp; i < stride; i++)
    					dst[i] = src[i] + ((prev[i] + dst[i]) >> 1);
    			    break;
    
    		    case 4: /* Paeth */
    				for (i = 0; i < bpp; i++)
    					dst[i] = src[i] + prev[i];
    				for (i = bpp; i < stride; i++)
    					dst[i] = src[i] + paeth(dst[i], prev[i], prev[i-bpp]);
    			    break;
    		}
        }
        src += stride;
        dst += stride;
	}
}

static const unsigned int adam7_ix[7] = { 0, 4, 0, 2, 0, 1, 0 };
static const unsigned int adam7_dx[7] = { 8, 8, 4, 4, 2, 2, 1 };
static const unsigned int adam7_iy[7] = { 0, 0, 4, 0, 2, 0, 1 };
static const unsigned int adam7_dy[7] = { 8, 8, 8, 4, 4, 2, 2 };

static void
png_deinterlace_passes(fz_context *ctx, struct info *info, unsigned int *w, unsigned int *h, unsigned int *ofs)
{
	int p, bpp = info->depth * info->n;
	ofs[0] = 0;
	for (p = 0; p < 7; p++)
	{
		w[p] = (info->width + adam7_dx[p] - adam7_ix[p] - 1) / adam7_dx[p];
		h[p] = (info->height + adam7_dy[p] - adam7_iy[p] - 1) / adam7_dy[p];
		if (w[p] == 0) h[p] = 0;
		if (h[p] == 0) w[p] = 0;
		if (w[p] && h[p])
			ofs[p + 1] = ofs[p] + h[p] * (1 + (w[p] * bpp + 7) / 8);
		else
			ofs[p + 1] = ofs[p];
	}
}

static void
png_deinterlace(fz_context *ctx, struct info *info, unsigned int *passw, unsigned int *passh, unsigned int *passofs)
{
	unsigned int n = info->n;
	unsigned int depth = info->depth;
	unsigned int stride = (info->width * n * depth + 7) / 8;
	unsigned char *output;
	unsigned int p, x, y, k;

	output = fz_malloc_array(ctx, info->height, stride);

	for (p = 0; p < 7; p++)
	{
		unsigned char *sp = info->samples + (passofs[p]);
		unsigned int w = passw[p];
		unsigned int h = passh[p];

		png_predict(sp, w, h, n, depth);
		for (y = 0; y < h; y++)
		{
			for (x = 0; x < w; x++)
			{
				int outx = x * adam7_dx[p] + adam7_ix[p];
				int outy = y * adam7_dy[p] + adam7_iy[p];
				unsigned char *dp = output + outy * stride;
				for (k = 0; k < n; k++)
				{
					int v = getcomp(sp, x * n + k, depth);
					putcomp(dp, outx * n + k, depth, v);
				}
			}
			sp += (w * depth * n + 7) / 8;
		}
	}

	free(info->samples);
	info->samples = output;
}

static int 
png_read_ihdr(fz_context *ctx, struct info *info, unsigned char *p, unsigned int size)
{
	int color, ret = 0;
    PNGHeader *hdr = (PNGHeader*)p;

	if (size != 13) {
	    fz_warn(ctx, "IHDR chunk is the wrong size");
	    return -1;
	}

	info->width = av_bswap32(hdr->width);
	info->height = av_bswap32(hdr->height);
	info->depth = hdr->bit_depth;

	color = hdr->color_type;
	info->interlace = hdr->interlace_type;

	if (info->depth != 1 && info->depth != 2 && info->depth != 4 && info->depth != 8 && info->depth != 16) {
		fz_warn(ctx, "image bit depth must be one of 1, 2, 4, 8, 16");
		ret = -1;
	}
	if (((color == 2) || (color == 4) || (color == 6)) && info->depth < 8) {
		fz_warn(ctx, "illegal bit depth for gray8/rgb24/rgb32a");
		ret = -1;
	}
	if (color == 3 && info->depth > 8) {
		fz_warn(ctx, "illegal bit depth for indexed");
		ret = -1;
	}

	info->indexed = 0;
	if (color == 0) /* gray */
		info->n = 1;
	else if (color == 2) /* rgb */
		info->n = 3;
	else if (color == 4) /* gray alpha */
		info->n = 2;
	else if (color == 6) /* rgb alpha */
		info->n = 4;
	else if (color == 3) /* indexed */
	{
		info->indexed = 1;
		info->n = 1;
	} else {
		fz_warn(ctx, "unknown color type");
		ret = -1;
	}

	if (info->interlace != 0 && info->interlace != 1) {
		fz_warn(ctx, "Interlace method not supported");
		ret = -1;
	}
	if (info->height > UINT_MAX / info->width / info->n / (info->depth / 8 + 1)) {
		fz_warn(ctx, "Image dimensions might overflow");
		ret = -1;
	}
	return ret;
}

static void
png_read_plte(fz_context *ctx, struct info *info, unsigned char *p, unsigned int size)
{
	int n = size / 3;
	int i;

	if (n > 256) {
		fz_warn(ctx, "too many samples in palette");
		n = 256;
	}

	for (i = 0; i < n; i++) {
		info->palette[i * 4] = p[i * 3];
		info->palette[i * 4 + 1] = p[i * 3 + 1];
		info->palette[i * 4 + 2] = p[i * 3 + 2];
	}

	/* Fill in any missing palette entries */
	for (; i < 256; i++) {
		info->palette[i * 4] = 0;
		info->palette[i * 4 + 1] = 0;
		info->palette[i * 4 + 2] = 0;
	}
}

static int
png_read_trns(fz_context *ctx, struct info *info, unsigned char *p, unsigned int size)
{
	unsigned int i;

	info->transparency = 1;

	if (info->indexed) {
		if (size > 256) {
			fz_warn(ctx, "too many samples in transparency table");
			size = 256;
		}
		for (i = 0; i < size; i++)
			info->palette[i * 4 + 3] = p[i];
		/* Fill in any missing entries */
		for (; i < 256; i++)
			info->palette[i * 4 + 3] = 255;
	} else {
		if (size != info->n * 2) {
			fz_warn(ctx, "tRNS chunk is the wrong size");
            return -1;
        }
		for (i = 0; i < info->n; i++)
			info->trns[i] = (p[i * 2] << 8 | p[i * 2 + 1]) & ((1 << info->depth) - 1);
	}
    return 0;
}

static int 
png_read_phys(fz_context *ctx, struct info *info, unsigned char *p, unsigned int size)
{
	if (size != 9) {
		fz_warn(ctx, "pHYs chunk is the wrong size");
        return -1;
    }
	if (p[8] == 1) {
		info->xres = (getuint(p) * 254 + 5000) / 10000;
		info->yres = (getuint(p + 4) * 254 + 5000) / 10000;
	}
    return 0;
}

static int 
png_read_image(fz_context *ctx, struct info *info, unsigned char *p, unsigned int total)
{
	unsigned int passw[7], passh[7], passofs[8];
	unsigned int code, size, i;
	z_stream stm;

	memset(info, 0, sizeof (struct info));
	info->xres = 96;
	info->yres = 96;

    for (i = 0; i < 256; i++) {
        info->palette[i] = 0xFF;
    }

	/* We assume the calling app already checked the signature, if not and it's not a PNG it's virtually guaranteed
	 * that we'll bail on the IHDR chunk test below anyway, so let's just skip right ahead to that. */
	p += 8;
	total -= 8;

	/* Read IHDR chunk (must come first) */
	size = getuint(p);
	if (total < 12 || size > total - 12) {
		fz_warn(ctx, "premature end of data in png image");
		return -1;
	}

	if (!memcmp(p + 4, "IHDR", 4)) {
	    if (size != 13) {
		    fz_warn(ctx, "IHDR chunk is the wrong size");
		    return -1;
	    }
		if (png_read_ihdr(ctx, info, p + 8, size) < 0)
			return -1;
	} else {
		fz_warn(ctx, "png file must start with IHDR chunk");
		return -1;
	}

	p += size + 12;
	total -= size + 12;

	/* Prepare output buffer */
	if (!info->interlace) {
		info->size = info->height * (1 + (info->width * info->n * info->depth + 7) / 8);
	} else {
		png_deinterlace_passes(ctx, info, passw, passh, passofs);
		info->size = passofs[7];
	}

	info->samples = fz_malloc(ctx, info->size);

	stm.zalloc = Z_NULL;
	stm.zfree = Z_NULL;
	stm.opaque = Z_NULL;

	stm.next_out = info->samples;
	stm.avail_out = info->size;

	code = inflateInit(&stm);
	if (code != Z_OK) {
		free(info->samples);
		fz_warn(ctx, "zlib error: %s", stm.msg);
		return -1;
	}

	/* Read remaining chunks until IEND */
	while (total > 8) {
		size = getuint(p);

		if (total < 12 || size > total - 12) {
			fz_warn(ctx, "premature end of data in png image");
		    inflateEnd(&stm);
		    free(info->samples);
            return -1;
        }

		if (!memcmp(p + 4, "PLTE", 4))
			png_read_plte(ctx, info, p + 8, size);
		if (!memcmp(p + 4, "tRNS", 4))
			png_read_trns(ctx, info, p + 8, size);
		if (!memcmp(p + 4, "pHYs", 4))
			png_read_phys(ctx, info, p + 8, size);
		if (!memcmp(p + 4, "IDAT", 4)) {
            int code;
	        stm.next_in = p + 8;
	        stm.avail_in = size;
	        code = inflate(&stm, Z_SYNC_FLUSH);
	        if (code != Z_OK && code != Z_STREAM_END) {
		        fz_warn(ctx, "zlib error: %s", stm.msg);
		        inflateEnd(&stm);
		        free(info->samples);
                return -1;
	        }
	        if (stm.avail_in != 0) {
		        if (stm.avail_out == 0) {
			        fz_warn(ctx, "ran out of output before input");
		            inflateEnd(&stm);
		            free(info->samples);
                    return -1;
                }
		        fz_warn(ctx, "inflate did not consume buffer (%u remaining)", stm.avail_in);
	        }
        }
		if (!memcmp(p + 4, "IEND", 4))
			break;

		p += size + 12;
		total -= size + 12;
	}
	if (stm.avail_out != 0) {
		fz_warn(ctx, "missing pixel data in png image; possibly truncated");
	} else if (total <= 8) {
		fz_warn(ctx, "missing IEND chunk in png image; possibly truncated");
	}

	code = inflateEnd(&stm);
	if (code != Z_OK) {
		free(info->samples);
		fz_warn(ctx, "zlib error: %s", stm.msg);
		return -1;
	}

	/* Apply prediction filter and deinterlacing */
	if (!info->interlace)
		png_predict(info->samples, info->width, info->height, info->n, info->depth);
	else
		png_deinterlace(ctx, info, passw, passh, passofs);
	return 0;
}

static fz_pixmap *
png_expand_palette(struct info *info, fz_pixmap *dst, fz_pixmap *src)
{
	unsigned char *sp = src->samples;
	unsigned char *dp = dst->samples;
	unsigned int x, y;

	dst->xres = src->xres;
	dst->yres = src->yres;

	for (y = info->height; y > 0; y--) {
		for (x = info->width; x > 0; x--) { 
			int v = *sp << 2;
			*dp++ = info->palette[v];
			*dp++ = info->palette[v + 1];
			*dp++ = info->palette[v + 2];
			*dp++ = info->palette[v + 3];
			sp += 2;
		}
	}
	return dst;
}

static void
png_mask_transparency(struct info *info, fz_pixmap *dst)
{
	unsigned int stride = (info->width * info->n * info->depth + 7) / 8;
	unsigned int depth = info->depth;
	unsigned int n = info->n;
	unsigned int x, y, k, t;

	for (y = 0; y < info->height; y++) {
		unsigned char *sp = info->samples + (unsigned int)(y * stride);
		unsigned char *dp = dst->samples + (unsigned int)(y * dst->w * dst->n);
		for (x = 0; x < info->width; x++) { 
			t = 1;
			for (k = 0; k < n; k++)
				if (getcomp(sp, x * n + k, depth) != info->trns[k])
					t = 0;
			if (t)
				dp[x * dst->n + dst->n - 1] = 0;
		}
	}
}

/* Unpack image samples */
#define get1(buf,x) ((buf[x >> 3] >> ( 7 - (x & 7) ) ) & 1 )
#define get2(buf,x) ((buf[x >> 2] >> ( ( 3 - (x & 3) ) << 1 ) ) & 3 )
#define get4(buf,x) ((buf[x >> 1] >> ( ( 1 - (x & 1) ) << 2 ) ) & 15 )
#define get8(buf,x) (buf[x])

static unsigned char get1_tab_1[256][8];
static unsigned char get1_tab_255[256][8];

static void init_get1_tables(void)
{
	static int once = 0;
	unsigned char bits[1];
	int i, k, x;

	if (once)
		return;

	for (i = 0; i < 256; i++) {
		bits[0] = i;
		for (k = 0; k < 8; k++) {
			x = get1(bits, k);
			get1_tab_1[i][k] = x;
			get1_tab_255[i][k] = x * 255;
		}
	}

	once = 1;
}

void fz_unpack_tile(fz_pixmap *dst, unsigned char * restrict src, int n, int depth, int stride, int scale)
{
	int pad = 0, x, y, k;
	int w = dst->w;

	if (dst->n > n)
		pad = 255;

	if (depth == 1)
		init_get1_tables();

	if (scale == 0) {
		switch (depth) {
		    case 1: scale = 255; break;
		    case 2: scale = 85; break;
		    case 4: scale = 17; break;
		}
	}

	for (y = 0; y < dst->h; y++) {
		unsigned char *sp = src + (unsigned int)(y * stride);
		unsigned char *dp = dst->samples + (unsigned int)(y * dst->w * dst->n);

		/* Specialized loops */
		if (n == 1 && depth == 1 && scale == 1) {
			int w3 = w >> 3;
			for (x = 0; x < w3; x++) {
				memcpy(dp, get1_tab_1[*sp++], 8);
				dp += 8;
			}
			x = x << 3;
			if (x < w)
				memcpy(dp, get1_tab_1[*sp], w - x);
		} else if (n == 1 && depth == 1 && scale == 255) {
			int w3 = w >> 3;
			for (x = 0; x < w3; x++) {
				memcpy(dp, get1_tab_255[*sp++], 8);
				dp += 8;
			}
			x = x << 3;
			if (x < w)
				memcpy(dp, get1_tab_255[*sp], w - x);
		} else if (depth == 8 && !pad) {
			int len = w * n;
			while (len--)
				*dp++ = *sp++;
		} else if (depth == 8 && pad) {
			for (x = 0; x < w; x++) {
				for (k = 0; k < n; k++)
					*dp++ = *sp++;
				*dp++ = 255;
			}
		} else {
			int b = 0;
			for (x = 0; x < w; x++) {
				for (k = 0; k < n; k++) {
					switch (depth) {
					    case 1: *dp++ = get1(sp, b) * scale; break;
					    case 2: *dp++ = get2(sp, b) * scale; break;
					    case 4: *dp++ = get4(sp, b) * scale; break;
					    case 8: *dp++ = get8(sp, b); break;
					}
					b++;
				}
				if (pad)
					*dp++ = 255;
			}
		}
	}
}

fz_pixmap *fz_load_png(fz_context *ctx, unsigned char *p, int total)
{
	fz_pixmap *image;
	fz_colorspace *colorspace;
	struct info png;
	int stride, r;

	r = png_read_image(ctx, &png, p, total);
	if (r < 0) {
		return NULL;
	}

	if (png.n == 3 || png.n == 4)
		colorspace = fz_device_rgb(ctx);
	else
		colorspace = fz_device_gray(ctx);

	stride = (png.width * png.n * png.depth + 7) / 8;

	image = fz_new_pixmap(ctx, colorspace, png.width, png.height);
	image->n = 1 + colorspace->n;
	image->xres = png.xres;
	image->yres = png.yres;

	fz_unpack_tile(image, png.samples, png.n, png.depth, stride, png.indexed);

	if (png.indexed) {
	    fz_pixmap *dst = fz_new_pixmap(ctx, fz_device_rgb(ctx), image->w, image->h);
		image = png_expand_palette(&png, dst, image);
	    fz_drop_pixmap(ctx, image);
	} else if (png.transparency) {
		png_mask_transparency(&png, image);
	}

	if (png.transparency || png.n == 2 || png.n == 4)
		fz_premultiply_pixmap(ctx, image);

	free(png.samples);
	return image;
}

void
fz_load_png_info(fz_context *ctx, unsigned char *p, int total, int *wp, int *hp, int *xresp, int *yresp, fz_colorspace **cspacep)
{
	struct info png;
	unsigned int size;

	/* Read IHDR chunk (must come first) */
	p += 8; total -= 8;
	size = getuint(p);
	if (total < 12 || size > total - 12) {
		fz_warn(ctx, "premature end of data in png image");
        return;
    }

	if (!memcmp(p + 4, "IHDR", 4)) {
		if (png_read_ihdr(ctx, &png, p + 8, size) < 0) {
            return;
        }
	} else {
		fz_warn(ctx, "png file must start with IHDR chunk");
        return;
    }

	if (png.n == 3 || png.n == 4)
		*cspacep = fz_device_rgb(ctx);
	else
		*cspacep = fz_device_gray(ctx);

	*wp = png.width;
	*hp = png.height;
	*xresp = 96;
	*yresp = 96;
}

