#include "mupdf/img.h"
#define DPI 72.0f

struct image_document_s
{
	fz_document super;
	fz_context *ctx;
    fz_buffer *buffer;
	fz_image *image;
};

image_page *
image_load_page(image_document *doc, int number)
{
	return (image_page *)doc->image;
}

void
image_free_page(image_document *doc, image_page *page)
{
}

fz_rect *
image_bound_page(image_document *doc, image_page *page, fz_rect *bbox)
{
	fz_image *image = (fz_image *)page;
	bbox->x0 = bbox->y0 = 0;
	bbox->x1 = image->w * DPI / image->xres;
	bbox->y1 = image->h * DPI / image->yres;
	return bbox;
}

void
image_run_page(image_document *doc, image_page *page, fz_device *dev, const fz_matrix *ctm, fz_cookie *cookie)
{
	fz_matrix local_ctm = *ctm;
	fz_image *image = (fz_image *)page;
	float w = image->w * DPI / image->xres;
	float h = image->h * DPI / image->yres;
	fz_pre_scale(&local_ctm, w, h);
	fz_fill_image(dev, image, &local_ctm, 1);
}

void
image_close_document(image_document *doc)
{
	fz_context *ctx = doc->ctx;
	fz_drop_image(ctx, doc->image);
	fz_drop_buffer(ctx, doc->buffer);
	fz_free(ctx, doc);
}

static image_document *
image_init_document(fz_context *ctx, fz_buffer *buf)
{
	image_document *doc = fz_malloc_struct(ctx, image_document);
	doc->super.close = (void*)image_close_document;
	doc->super.load_page = (void*)image_load_page;
	doc->super.bound_page = (void*)image_bound_page;
	doc->super.run_page_contents = (void*)image_run_page;
	doc->super.free_page = (void*)image_free_page;
	doc->super.meta = (void*)image_meta;
	doc->ctx = ctx;
    doc->buf = fz_keep_buffer(buf);

	fz_try(ctx)
	{
		doc->image = fz_new_image_from_buffer(ctx, buffer);
	}
	fz_catch(ctx)
	{
		image_close_document(doc);
		fz_rethrow(ctx);
	}
}

image_document *
image_open_document(fz_context *ctx, unsigned char *data, unsigned int datasize)
{
    fz_buffer *buf = fz_new_buffer_from_data(ctx, data, datasize);
	return image_init_document(ctx, buf);
}

