#include "mupdf/pdf.h"
#include "cmaps_core.h"

/* Katakana */
static const pdf_range cmap_Katakana_ranges[] = {
{32,253,326},{96,121,391},
};

static pdf_cmap cmap_Katakana = {
	{-1, pdf_free_cmap_imp}, "Katakana", "", NULL, 0, 1,
	{ {1,0,255}, },
	2, 2, (pdf_range*) cmap_Katakana_ranges,
	0, 0, NULL,
	0, 0, NULL,
};

/* Roman */
static const pdf_range cmap_Roman_ranges[] = {
{32,377,231},
};

static pdf_cmap cmap_Roman = {
	{-1, pdf_free_cmap_imp}, "Roman", "", NULL, 0, 1,
	{ {1,0,255}, },
	1, 1, (pdf_range*) cmap_Roman_ranges,
	0, 0, NULL,
	0, 0, NULL,
};

/* Adobe-GB1-UCS2 */
static pdf_cmap cmap_Adobe_GB1_UCS2 = {
	{-1, pdf_free_cmap_imp}, "Adobe-GB1-UCS2", "", NULL, 0, 1,
	{ {2,0,65535}, },
	4383, 4383, (pdf_range*) cmap_Adobe_GB1_UCS2_ranges,
	9975, 9975, (unsigned short*) cmap_Adobe_GB1_UCS2_table,
	0, 0, NULL,
};

/* Adobe-CNS1-UCS2 */
static pdf_cmap cmap_Adobe_CNS1_UCS2 = {
	{-1, pdf_free_cmap_imp}, "Adobe-CNS1-UCS2", "", NULL, 0, 1,
	{ {2,0,65535}, },
	1184, 1184, (pdf_range*) cmap_Adobe_CNS1_UCS2_ranges,
	16937, 16937, (unsigned short*) cmap_Adobe_CNS1_UCS2_table,
	0, 0, NULL,
};

/* Adobe-Japan1-UCS2 */
static pdf_cmap cmap_Adobe_Japan1_UCS2 = {
	{-1, pdf_free_cmap_imp}, "Adobe-Japan1-UCS2", "", NULL, 0, 1,
	{ {2,0,65535}, },
	2104, 2104, (pdf_range*) cmap_Adobe_Japan1_UCS2_ranges,
	15838, 15838, (unsigned short*) cmap_Adobe_Japan1_UCS2_table,
	0, 0, NULL,
};

/* Adobe-Japan2-0 */
static const pdf_range cmap_Adobe_Japan2_0_ranges[] = {
{0,24269,0},
};

static pdf_cmap cmap_Adobe_Japan2_0 = {
	{-1, pdf_free_cmap_imp}, "Adobe-Japan2-UCS2", "", NULL, 0, 1,
	{ {2,0,6143}, },
	1, 1, (pdf_range*) cmap_Adobe_Japan2_0_ranges,
	0, 0, NULL,
	0, 0, NULL,
};

/* Identity-H */

static const pdf_range cmap_Identity_H_ranges[] = {
  {0x0000,0xfffd,0},{0x3fff,0xfffd,16383},{0x7ffe,0xfffd,32766},{0xbffd,0xfffd,49149},
  {0xfffc,0x0d,65532},
};

static pdf_cmap cmap_Identity1_H = {
	{-1, pdf_free_cmap_imp}, "Identity1-H", "", 0, 0,
	1, { {1,0x0,0xffff}, },
	5, 5, (pdf_range*) cmap_Identity_H_ranges,
	0, 0, (unsigned short*) NULL,
	0, 0, (pdf_xrange*) NULL,
};

static pdf_cmap cmap_Identity2_H = {
	{-1, pdf_free_cmap_imp}, "Identity2-H", "", 0, 0,
	1, { {2,0x0,0xffff}, },
	5, 5, (pdf_range*) cmap_Identity_H_ranges,
	0, 0, (unsigned short*) NULL,
	0, 0, (pdf_xrange*) NULL,
};

static const struct { char *name; pdf_cmap *cmap; } cmap_table[] =
{
	{"Adobe-CNS1-UCS2",&cmap_Adobe_CNS1_UCS2},
	{"Adobe-GB1-UCS2",&cmap_Adobe_GB1_UCS2},
	{"Adobe-Japan1-UCS2",&cmap_Adobe_Japan1_UCS2},
	{"Adobe-Japan2-0",&cmap_Adobe_Japan2_0},
	{"Katakana",&cmap_Katakana},
	{"Roman",&cmap_Roman},
};

pdf_cmap *
pdf_load_builtin_cmap(fz_context *ctx, char *cmap_name)
{
	int l = 0;
	size_t r = nelem(cmap_table) - 1;
	while (l <= r)
	{
		int m = (l + r) >> 1;
		int c = strcmp(cmap_name, cmap_table[m].name);
		if (c < 0)
			r = m - 1;
		else if (c > 0)
			l = m + 1;
		else
			return cmap_table[m].cmap;
	}
	return NULL;
}

